--- old/usr/src/lib/libc/port/locale/rune.c	Sun Jun  8 22:09:55 2014
+++ new/usr/src/lib/libc/port/locale/rune.c	Sun Jun  8 22:09:55 2014
@@ -1,4 +1,5 @@
 /*
+ * Copyright 2014 Garrett D'Amore <garrett@damore.org>
  * Copyright 2010 Nexenta Systems, Inc.  All rights reserved.
  * Copyright (c) 1993
  *	The Regents of the University of California.  All rights reserved.
@@ -39,14 +40,15 @@
 #include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/mman.h>
+#include <fcntl.h>
+#include <unistd.h>
 
 #include "runetype.h"
 #include "runefile.h"
 
-_RuneLocale *_Read_RuneMagi(FILE *);
-
 _RuneLocale *
-_Read_RuneMagi(FILE *fp)
+_Read_RuneMagi(const char *fname)
 {
 	char *fdata, *data;
 	void *lastp;
@@ -61,31 +63,30 @@
 	_FileRuneEntry *maplower_ext_ranges;
 	_FileRuneEntry *mapupper_ext_ranges;
 	int runetype_ext_len = 0;
+	int fd;
 
-	if (fstat(fileno(fp), &sb) < 0)
+	if ((fd = open(fname, O_RDONLY)) < 0) {
+		errno = EINVAL;
 		return (NULL);
+	}
 
-	if ((size_t)sb.st_size < sizeof (_FileRuneLocale)) {
+	if (fstat(fd, &sb) < 0) {
+		(void) close(fd);
 		errno = EINVAL;
 		return (NULL);
 	}
 
-	if ((fdata = malloc(sb.st_size)) == NULL)
+	if ((size_t)sb.st_size < sizeof (_FileRuneLocale)) {
+		(void) close(fd);
+		errno = EINVAL;
 		return (NULL);
-
-	errno = 0;
-	rewind(fp); /* Someone might have read the magic number once already */
-	if (errno) {
-		saverr = errno;
-		free(fdata);
-		errno = saverr;
-		return (NULL);
 	}
 
-	if (fread(fdata, sb.st_size, 1, fp) != 1) {
-		saverr = errno;
-		free(fdata);
-		errno = saverr;
+
+	fdata = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
+	(void) close(fd);
+	if (fdata == NULL) {
+		errno = EINVAL;
 		return (NULL);
 	}
 
@@ -95,33 +96,25 @@
 	variable = frl + 1;
 
 	if (memcmp(frl->magic, _FILE_RUNE_MAGIC_1, sizeof (frl->magic))) {
-		free(fdata);
-		errno = EINVAL;
-		return (NULL);
+		goto invalid;
 	}
 
 	runetype_ext_ranges = (_FileRuneEntry *)variable;
 	variable = runetype_ext_ranges + frl->runetype_ext_nranges;
 	if (variable > lastp) {
-		free(fdata);
-		errno = EINVAL;
-		return (NULL);
+		goto invalid;
 	}
 
 	maplower_ext_ranges = (_FileRuneEntry *)variable;
 	variable = maplower_ext_ranges + frl->maplower_ext_nranges;
 	if (variable > lastp) {
-		free(fdata);
-		errno = EINVAL;
-		return (NULL);
+		goto invalid;
 	}
 
 	mapupper_ext_ranges = (_FileRuneEntry *)variable;
 	variable = mapupper_ext_ranges + frl->mapupper_ext_nranges;
 	if (variable > lastp) {
-		free(fdata);
-		errno = EINVAL;
-		return (NULL);
+		goto invalid;
 	}
 
 	frr = runetype_ext_ranges;
@@ -134,17 +127,13 @@
 			variable = types + len;
 			runetype_ext_len += len;
 			if (variable > lastp) {
-				free(fdata);
-				errno = EINVAL;
-				return (NULL);
+				goto invalid;
 			}
 		}
 	}
 
 	if ((char *)variable + frl->variable_len > (char *)lastp) {
-		free(fdata);
-		errno = EINVAL;
-		return (NULL);
+		goto invalid;
 	}
 
 	/*
@@ -157,7 +146,7 @@
 	    frl->variable_len);
 	if (data == NULL) {
 		saverr = errno;
-		free(fdata);
+		(void) munmap(fdata, sb.st_size);
 		errno = saverr;
 		return (NULL);
 	}
@@ -229,7 +218,7 @@
 	}
 
 	(void) memcpy(rl->__variable, variable, rl->__variable_len);
-	free(fdata);
+	(void) munmap(fdata, sb.st_size);
 
 	/*
 	 * Go out and zero pointers that should be zero.
@@ -247,4 +236,9 @@
 		rl->__mapupper_ext.__ranges = NULL;
 
 	return (rl);
+
+invalid:
+	(void) munmap(fdata, sb.st_size);
+	errno = EINVAL;
+	return (NULL);
 }
